"""Contains all non-interacting functionality and solvers."""


import copy
import itertools
from tqdm import tqdm
from collections.abc import Callable
import numpy as np
import scipy.sparse as sps
import numpy.linalg as npla
import scipy.linalg as spla
import scipy.sparse.linalg as spsla
import iDEA.system
import iDEA.state
import iDEA.observables


name = "non_interacting"


def kinetic_energy_operator(s: iDEA.system.System) -> np.ndarray:
    r"""
    Compute single-particle kinetic energy operator as a matrix.

    This is built using a given number of finite differences to represent the second derivative.
    The number of differences taken is defined in s.stencil.

    | Args:
    |     s: iDEA.system.System, System object.

    | Returns:
    |     K: np.ndarray, Kintetic energy operator.
    """
    if s.stencil == 3:
        sd = 1.0 * np.array([1, -2, 1], dtype=np.float) / s.dx**2
        sdi = (-1, 0, 1)
    elif s.stencil == 5:
        sd = 1.0 / 12.0 * np.array([-1, 16, -30, 16, -1], dtype=np.float) / s.dx**2
        sdi = (-2, -1, 0, 1, 2)
    elif s.stencil == 7:
        sd = (
            1.0
            / 180.0
            * np.array([2, -27, 270, -490, 270, -27, 2], dtype=np.float)
            / s.dx**2
        )
        sdi = (-3, -2, -1, 0, 1, 2, 3)
    elif s.stencil == 9:
        sd = (
            1.0
            / 5040.0
            * np.array(
                [-9, 128, -1008, 8064, -14350, 8064, -1008, 128, -9], dtype=np.float
            )
            / s.dx**2
        )
        sdi = (-4, -3, -2, -1, 0, 1, 2, 3, 4)
    elif s.stencil == 11:
        sd = (
            1.0
            / 25200.0
            * np.array(
                [8, -125, 1000, -6000, 42000, -73766, 42000, -6000, 1000, -125, 8],
                dtype=np.float,
            )
            / s.dx**2
        )
        sdi = (-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5)
    elif s.stencil == 13:
        sd = (
            1.0
            / 831600.0
            * np.array(
                [
                    -50,
                    864,
                    -7425,
                    44000,
                    -222750,
                    1425600,
                    -2480478,
                    1425600,
                    -222750,
                    44000,
                    -7425,
                    864,
                    -50,
                ],
                dtype=np.float,
            )
            / s.dx**2
        )
        sdi = (-6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6)
    second_derivative = np.zeros((s.x.shape[0], s.x.shape[0]))
    for i in range(len(sdi)):
        second_derivative += np.diag(
            np.full(
                np.diag(np.zeros((s.x.shape[0], s.x.shape[0])), k=sdi[i]).shape[0],
                sd[i],
            ),
            k=sdi[i],
        )
    K = -0.5 * second_derivative
    return K


def external_potential_operator(s: iDEA.system.System) -> np.ndarray:
    r"""
    Compute the external potential operator.

    | Args:
    |     s: iDEA.system.System, System object.

    | Returns:
    |     Vext: np.ndarray, External potential energy operator.
    """
    Vext = np.diag(s.v_ext)
    return Vext


def hamiltonian(
    s: iDEA.system.System,
    up_n: np.ndarray = None,
    down_n: np.ndarray = None,
    up_p: np.ndarray = None,
    down_p: np.ndarray = None,
    K: np.ndarray = None,
    Vext: np.ndarray = None,
) -> np.ndarray:
    r"""
    Compute the Hamiltonian from the kinetic and potential terms.

    | Args:
    |     s: iDEA.system.System, System object.
    |     up_n: np.ndarray, Charge density of up electrons.
    |     down_n: np.ndarray, Charge density of down electrons.
    |     up_p: np.ndarray, Charge density matrix of up electrons.
    |     down_p: np.ndarray, Charge density matrix of down electrons.
    |     K: np.ndarray, Single-particle kinetic energy operator [If None this will be computed from s]. (default = None)
    |     Vext: np.ndarray, Potential energy operator [If None this will be computed from s]. (default = None)

    | Returns:
    |     H: np.ndarray, Hamiltonian, up Hamiltonian, down Hamiltonian.
    """
    if K is None:
        K = kinetic_energy_operator(s)
    if Vext is None:
        Vext = external_potential_operator(s)
    H = K + Vext
    return H, H, H


def total_energy(s: iDEA.system.System, state: iDEA.state.SingleBodyState) -> float:
    r"""
    Compute the total energy of a non_interacting state.

    | Args:
    |     s: iDEA.system.System, System object.
    |     state: iDEA.state.SingleBodyState, State. (default = None)

    | Returns:
    |     E: float, Total energy.
    """
    return iDEA.observables.single_particle_energy(s, state)


def add_occupations(
    s: iDEA.system.System, state: iDEA.state.SingleBodyState, k: int
) -> iDEA.state.SingleBodyState:
    r"""
    Calculate the occpuations of a state in a given energy excitation.

    | Args:
    |     s: iDEA.system.System, System object.
    |     state: iDEA.state.SingleBodyState, State.
    |     k: int, Excitation state [k = 0 is the ground-state].

    | Returns:
    |     state: iDEA.state.SingleBodyState, State with occupations added.
    """
    # Calculate the max level or orbitals needed to achieve required state and only use these.
    max_level = (k + 1) * int(np.ceil(s.count))
    up_energies = state.up.energies[:max_level]
    down_energies = state.down.energies[:max_level]

    # Calculate all possible combinations of spin indices.
    up_indices = list(itertools.combinations(range(max_level), s.up_count))
    down_indices = list(itertools.combinations(range(max_level), s.down_count))

    # Construct all possible occupations.
    up_occupations = []
    for up_index in up_indices:
        up_occupation = np.zeros(shape=up_energies.shape, dtype=np.float)
        np.put(up_occupation, up_index, [1.0] * s.up_count)
        up_occupations.append(up_occupation)
    down_occupations = []
    for down_index in down_indices:
        down_occupation = np.zeros(shape=down_energies.shape, dtype=np.float)
        np.put(down_occupation, down_index, [1.0] * s.down_count)
        down_occupations.append(down_occupation)
    occupations = list(itertools.product(up_occupations, down_occupations))

    # Calculate the energies of all possible occpuations.
    energies = []
    for occupation in occupations:
        state_copy = copy.deepcopy(state)
        state_copy.up.occupations = np.zeros(shape=state_copy.up.energies.shape)
        state_copy.up.occupations[:max_level] = occupation[0]
        state_copy.down.occupations = np.zeros(shape=state_copy.down.energies.shape)
        state_copy.down.occupations[:max_level] = occupation[1]
        E = iDEA.observables.single_particle_energy(s, state_copy)
        energies.append(E)

    # Choose the correct energy index.
    energy_index = np.argsort(energies)[k]

    # Construct correct occupations.
    state.up.occupations = np.zeros(shape=state_copy.up.energies.shape)
    state.up.occupations[:max_level] = occupations[energy_index][0]
    state.up.occupied = np.nonzero(state.up.occupations)[0]
    state.down.occupations = np.zeros(shape=state_copy.down.energies.shape)
    state.down.occupations[:max_level] = occupations[energy_index][1]
    state.down.occupied = np.nonzero(state.down.occupations)[0]

    return state


def sc_step(
    s: iDEA.system.System,
    state: iDEA.state.SingleBodyState,
    up_H: np.ndarray,
    down_H: np.ndarray,
):
    r"""
    Performs a single step of the self-consistent cycle.

    | Args:
    |     s: iDEA.system.System, System object.
    |     state: iDEA.state.SingleBodyState, Previous state.
    |     up_H: np.ndarray, Hamiltonian for up electrons.
    |     down_H: np.ndarray, Hamiltonian for down electrons.

    | Returns:
    |     state: iDEA.state.SingleBodyState, New state.
    """
    # Solve the non-interacting Schrodinger equation.
    state.up.energies, state.up.orbitals = spla.eigh(up_H)
    state.down.energies, state.down.orbitals = spla.eigh(down_H)

    # Normalise orbitals.
    state.up.orbitals = state.up.orbitals / np.sqrt(s.dx)
    state.down.orbitals = state.down.orbitals / np.sqrt(s.dx)

    return state


def solve(
    s: iDEA.system.System,
    hamiltonian_function: Callable = None,
    k: int = 0,
    restricted: bool = False,
    mixing: float = 0.5,
    tol: float = 1e-10,
    initial: tuple = None,
    name: str = "non_interacting",
    silent: bool = False,
    **kwargs
) -> iDEA.state.SingleBodyState:
    r"""
    Solves the Schrodinger equation for the given system.

    | Args:
    |     s: iDEA.system.System, System object.
    |     hamiltonian_function: Callable, Hamiltonian function [If None this will be the non_interacting function]. (default = None)
    |     k: int, Energy state to solve for. (default = 0, the ground-state)
    |     restricted: bool, Is the calculation restricted (r) on unrestricted (u). (default=False)
    |     mixing: float, Mixing parameter. (default = 0.5)
    |     tol: float, Tollerance of convergence. (default = 1e-10)
    |     initial: tuple. Tuple of initial values used to begin the self-consistency (n, up_n, down_n, p, up_p, down_p). (default = None)
    |     name: str, Name of method. (default = "non_interacting")
    |     silent: bool, Set to true to prevent printing. (default = False)


    | Returns:
    |     state: iDEA.state.SingleBodyState, Solved state.
    """
    # Construct the single-body state.
    state = iDEA.state.SingleBodyState()
    state.up.occupations = np.zeros(shape=s.x.shape)
    state.up.occupations[: s.up_count] = 1.0
    state.down.occupations = np.zeros(shape=s.x.shape)
    state.down.occupations[: s.down_count] = 1.0

    # Determine the Hamiltonian function.
    if hamiltonian_function is None:
        hamiltonian_function = hamiltonian

    # Construct the initial values.
    if initial is None:
        n_old = np.zeros(shape=s.x.shape)
        up_n_old = np.zeros(shape=s.x.shape)
        down_n_old = np.zeros(shape=s.x.shape)
        p_old = np.zeros(shape=s.x.shape * 2)
        up_p_old = np.zeros(shape=s.x.shape * 2)
        down_p_old = np.zeros(shape=s.x.shape * 2)
    else:
        n_old = initial[0]
        up_n_old = initial[1]
        down_n_old = initial[2]
        p_old = initial[3]
        up_p_old = initial[4]
        down_p_old = initial[5]

    # Construct the initial Hamiltonian. (And break the symmetry.)
    H_old, up_H_old, down_H_old = hamiltonian_function(
        s, up_n_old, down_n_old, up_p_old, down_p_old, **kwargs
    )
    H, up_H, down_H = hamiltonian_function(
        s, up_n_old, down_n_old, up_p_old, down_p_old, **kwargs
    )
    down_H += sps.spdiags(
        1e-12 * s.x, np.array([0]), s.x.shape[0], s.x.shape[0]
    ).toarray()

    # Apply restriction.
    if restricted:
        up_H_old = H_old
        down_H_old = H_old
        up_H = H
        down_H = H

    # Run self-consitent algorithm.
    convergence = 1.0
    count = 0
    while convergence > tol:

        # Perform single self-consistent step.
        state = sc_step(s, state, up_H, down_H)

        # Update values.
        n, up_n, down_n = iDEA.observables.density(s, state, return_spins=True)
        p, up_p, down_p = iDEA.observables.density_matrix(s, state, return_spins=True)

        # Perform mixing.
        if count != 0:
            n = mixing * n + (1.0 - mixing) * n_old
            up_n = mixing * up_n + (1.0 - mixing) * up_n_old
            down_n = mixing * down_n + (1.0 - mixing) * down_n_old
            p = mixing * p + (1.0 - mixing) * p_old
            up_p = mixing * up_p + (1.0 - mixing) * up_p_old
            down_p = mixing * down_p + (1.0 - mixing) * down_p_old

        # Construct the new Hamiltonian.
        H, up_H, down_H = hamiltonian_function(s, up_n, down_n, up_p, down_p, **kwargs)

        # Apply restriction.
        if restricted:
            up_H_old = H_old
            down_H_old = H_old
            up_H = H
            down_H = H

        # Test for convergence.
        convergence = np.sum(abs(n - n_old)) * s.dx

        # Update old values.
        n_old = n
        up_n_old = up_n
        down_n_old = down_n
        p_old = p
        up_p_old = up_p
        down_p_old = down_p

        # Update terminal.
        count += 1
        if silent is False:
            print(
                r"iDEA.methods.{0}.solve: convergence = {1:.5}, tollerance = {2:.5}".format(
                    name, convergence, tol
                ),
                end="\r",
            )

    # Compute state occupations.
    state = add_occupations(s, state, k)
    if silent is False:
        print()

    return state


def propagate_step(
    s: iDEA.system.System,
    evolution: iDEA.state.SingleBodyEvolution,
    j: int,
    hamiltonian_function: Callable,
    v_ptrb: np.ndarray,
    dt: float,
    restricted: bool,
    **kwargs
):
    r"""
    Propagate a set of orbitals forward in time due to a dynamic local pertubation.

    | Args:
    |     s: iDEA.system.System, System object.
    |     evolution: iDEA.state.SingleBodyEvolution, Time-dependent evolution.
    |     j: int, Time index to step to.
    |     hamiltonian_function: Callable, Hamiltonian function [If None this will be the non_interacting function]. (default = None)
    |     v_ptrb: np.ndarray, Local perturbing potential on the grid of t and x values, indexed as v_ptrb[time,space].
    |     dt: float, Timestep.
    |     restricted: bool, Is the calculation restricted (r) on unrestricted (u). (default=False)

    | Returns:
    |     evolution: iDEA.state.SingleBodyEvolution, Time-dependent evolution solved at time index j from j-1.
    """
    n, up_n, down_n = iDEA.observables.density(
        s, evolution=evolution, time_indices=np.array([j - 1]), return_spins=True
    )
    p, up_p, down_p = iDEA.observables.density_matrix(
        s, evolution=evolution, time_indices=np.array([j - 1]), return_spins=True
    )
    H, up_H, down_H = hamiltonian_function(
        s, up_n[0, ...], down_n[0, ...], up_p[0, ...], down_p[0, ...], **kwargs
    )
    H = sps.csc_matrix(H)
    up_H = sps.csc_matrix(up_H)
    down_H = sps.csc_matrix(down_H)
    Vptrb = sps.diags(v_ptrb[j, :]).tocsc()

    # Apply restriction.
    if restricted:
        up_H = H
        down_H = H

    for i in range(evolution.up.occupied.shape[0]):
        up_O = -1.0j * (up_H + Vptrb) * dt
        evolution.up.td_orbitals[j, :, i] = spsla.expm_multiply(
            up_O, evolution.up.td_orbitals[j - 1, :, i]
        )
        norm = npla.norm(evolution.up.td_orbitals[j, :, i]) * np.sqrt(s.dx)
        evolution.up.td_orbitals[j, :, i] /= norm
    for i in range(evolution.down.occupied.shape[0]):
        down_O = -1.0j * (down_H + Vptrb) * dt
        evolution.down.td_orbitals[j, :, i] = spsla.expm_multiply(
            down_O, evolution.down.td_orbitals[j - 1, :, i]
        )
        norm = npla.norm(evolution.down.td_orbitals[j, :, i]) * np.sqrt(s.dx)
        evolution.down.td_orbitals[j, :, i] /= norm

    return evolution


def propagate(
    s: iDEA.system.System,
    state: iDEA.state.SingleBodyState,
    v_ptrb: np.ndarray,
    t: np.ndarray,
    hamiltonian_function: Callable = None,
    restricted: bool = False,
    name: str = "non_interacting",
    **kwargs
) -> iDEA.state.SingleBodyEvolution:
    r"""
    Propagate a set of orbitals forward in time due to a dynamic local pertubation.

    | Args:
    |    s: iDEA.system.System, System object.  
    |    state: iDEA.state.SingleBodyState, State to be propigated. 
    |    v_ptrb: np.ndarray, Local perturbing potential on the grid of t and x values, indexed as v_ptrb[time,space]. 
    |    t: np.ndarray, Grid of time values. \n
    |    hamiltonian_function: Callable, Hamiltonian function [If None this will be the non_interacting function]. (default = None) 
    |    restricted: bool, Is the calculation restricted (r) on unrestricted (u). (default=False) 
    |    name: str, Name of method. (default = "non_interacting") 

    | Returns:
    |    evolution: iDEA.state.SingleBodyEvolution, Solved time-dependent evolution.
    """
    # Determine the Hamiltonian function.
    if hamiltonian_function is None:
        hamiltonian_function = hamiltonian

    # Construct the unperturbed Hamiltonian.
    n, up_n, down_n = iDEA.observables.density(s, state=state, return_spins=True)
    p, up_p, down_p = iDEA.observables.density_matrix(s, state=state, return_spins=True)
    H, up_H, down_H = hamiltonian_function(s, up_n, down_n, up_p, down_p, **kwargs)
    H = sps.csc_matrix(H)
    up_H = sps.csc_matrix(up_H)
    down_H = sps.csc_matrix(down_H)
    down_H += sps.spdiags(1e-12 * s.x, np.array([0]), s.x.shape[0], s.x.shape[0])

    # Apply restriction.
    if restricted:
        up_H = H
        down_H = H

    # Compute timestep.
    dt = t[1] - t[0]

    # Initilise the single-body time-dependent evolution.
    evolution = iDEA.state.SingleBodyEvolution(state)
    evolution.up.td_orbitals = np.zeros(
        shape=(t.shape[0], s.x.shape[0], state.up.occupied.shape[0]), dtype=np.complex
    )
    evolution.down.td_orbitals = np.zeros(
        shape=(t.shape[0], s.x.shape[0], state.down.occupied.shape[0]), dtype=np.complex
    )
    evolution.up.td_orbitals[0, :, :] = state.up.orbitals[:, state.up.occupied]
    evolution.down.td_orbitals[0, :, :] = state.down.orbitals[:, state.down.occupied]
    evolution.v_ptrb = v_ptrb
    evolution.t = t

    # Propagate.
    for j, ti in enumerate(
        tqdm(t, desc="iDEA.methods.{}.propagate: propagating state".format(name))
    ):
        if j != 0:
            evolution = propagate_step(
                s, evolution, j, hamiltonian_function, v_ptrb, dt, restricted, **kwargs
            )

    return evolution



import copy
import string
import itertools
from typing import Union
import numpy as np
import iDEA.system
import iDEA.state
import iDEA.methods.non_interacting
import iDEA.methods.interacting


def observable(
    s: iDEA.system.System,
    observable_operator: np.ndarray,
    state: Union[iDEA.state.SingleBodyState, iDEA.state.ManyBodyState] = None,
    evolution: Union[
        iDEA.state.SingleBodyEvolution, iDEA.state.ManyBodyEvolution
    ] = None,
    return_spins: bool = False,
) -> Union[float, np.ndarray]:
    r"""
    Compute an observable based on a given operator and state or evolution.

    | Args:
    |     s: iDEA.system.System, System object.
    |     observable_operator: np.ndarray, Obserbable operator.
    |     state: iDEA.state.SingleBodyState or iDEA.state.ManyBodyState, State. (default = None)
    |     evolution: iDEA.state.SingleBodyEvolution or iDEA.state.ManyBodyEvolution, Evolution. (default = None)
    |     return_spins: bool, True to also return the spin observables: total, up, down. (default = False)

    | Returns:
    |     observable: float or np.ndarray, Observable.
    """
    if state is not None and type(state) == iDEA.state.ManyBodyState:
        raise NotImplementedError()

    if state is not None and type(state) == iDEA.state.SingleBodyState:
        up_O = 0.0
        for i in range(state.up.orbitals.shape[1]):
            up_O += (
                np.vdot(
                    state.up.orbitals[:, i],
                    np.dot(observable_operator, state.up.orbitals[:, i]),
                )
                * state.up.occupations[i]
                * s.dx
            )
        down_O = 0.0
        for i in range(state.down.orbitals.shape[1]):
            down_O += (
                np.vdot(
                    state.down.orbitals[:, i],
                    np.dot(observable_operator, state.down.orbitals[:, i]),
                )
                * state.down.occupations[i]
                * s.dx
            )
        O = up_O + down_O
        if return_spins:
            return O, up_O, down_O
        else:
            return O

    if evolution is not None and type(evolution) == iDEA.state.ManyBodyEvolution:
        raise NotImplementedError()

    if evolution is not None and type(evolution) == iDEA.state.SingleBodyEvolution:
        up_O = np.zeros(shape=evolution.t.shape, dtype=complex)
        for i, I in enumerate(evolution.up.occupied):
            for j, ti in enumerate(evolution.t):
                up_O[j] += (
                    np.vdot(
                        evolution.up.td_orbitals[j, :, i],
                        np.dot(observable_operator, evolution.up.td_orbitals[j, :, i]),
                    )
                    * evolution.up.occupations[I]
                    * s.dx
                )
        down_O = np.zeros(shape=evolution.t.shape, dtype=complex)
        for i, I in enumerate(evolution.down.occupied):
            for j, ti in enumerate(evolution.t):
                down_O[j] += (
                    np.vdot(
                        evolution.down.td_orbitals[j, :, i],
                        np.dot(
                            observable_operator, evolution.down.td_orbitals[j, :, i]
                        ),
                    )
                    * evolution.down.occupations[I]
                    * s.dx
                )
        O = up_O + down_O
        if return_spins:
            return O.real, up_O.real, down_O.real
        else:
            return O.real

    else:
        raise AttributeError(f"State or Evolution must be provided.")


def density(
    s: iDEA.system.System,
    state: Union[iDEA.state.SingleBodyState, iDEA.state.ManyBodyState] = None,
    evolution: Union[
        iDEA.state.SingleBodyEvolution, iDEA.state.ManyBodyEvolution
    ] = None,
    time_indices: np.ndarray = None,
    return_spins: bool = False,
) -> np.ndarray:
    r"""
    Compute the charge density of a non_interacting state.

    | Args:
    |     s: iDEA.system.System, System object.
    |     state: iDEA.state.SingleBodyState or iDEA.state.ManyBodyState, State. (default = None)
    |     evolution: iDEA.state.SingleBodyEvolution or iDEA.state.ManyBodyEvolution, Evolution. (default = None)
    |     time_indices: np.ndarray, Time indices to compute observable if given evolution. If None will perform for all time indices. (default = None)
    |     return_spins: bool, True to also return the spin densities: total, up, down. (default = False)

    | Returns:
    |     n: np.ndarray, Charge density, or evolution of charge density.
    """
    if state is not None and type(state) == iDEA.state.ManyBodyState:
        spin_densities = np.zeros(shape=(s.x.shape[0], 2))
        for i in range(s.x.shape[0]):
            for k in range(2):
                spin_densities[i, k] = (
                    np.sum(abs(state.full[i, k, ...]) ** 2)
                    * s.dx ** (s.count - 1)
                    * s.count
                )
        up_n = spin_densities[:, 0]
        down_n = spin_densities[:, 1]
        n = up_n + down_n
        if return_spins:
            return n, up_n, down_n
        else:
            return n

    if state is not None and type(state) == iDEA.state.SingleBodyState:
        up_n = np.zeros(shape=s.x.shape[0])
        down_n = np.zeros(shape=s.x.shape[0])
        for i in range(state.up.orbitals.shape[1]):
            up_n += abs(state.up.orbitals[:, i]) ** 2 * state.up.occupations[i]
        for i in range(state.down.orbitals.shape[1]):
            down_n += abs(state.down.orbitals[:, i]) ** 2 * state.down.occupations[i]
        n = up_n + down_n
        if return_spins:
            return n, up_n, down_n
        else:
            return n

    if evolution is not None and type(evolution) == iDEA.state.ManyBodyEvolution:
        if time_indices is None:
            time_indices = np.array(range(evolution.t.shape[0]))
        spin_densities = np.zeros(shape=(time_indices.shape[0], s.x.shape[0], 2))
        for j, ti in enumerate(time_indices):
            l = string.ascii_lowercase[: s.count]
            L = string.ascii_uppercase[: s.count]
            st = (
                l
                + ","
                + L
                + "->"
                + "".join([i for sub in list(zip(l, L)) for i in sub])
            )
            full = np.einsum(st, evolution.td_space[ti, ...], evolution.spin)
            L = list(
                zip(list(range(0, s.count * 2, 2)), list(range(1, s.count * 2, 2)))
            )
            perms = itertools.permutations(list(range(s.count)))
            full_copy = copy.deepcopy(full)
            full = np.zeros_like(full_copy)
            for p in perms:
                indices = list(itertools.chain(*[L[e] for e in p]))
                full += iDEA.methods.interacting._permutation_parity(p) * np.moveaxis(
                    full_copy, list(range(s.count * 2)), indices
                )
            full = full / np.sqrt(np.sum(abs(full) ** 2) * s.dx**s.count)
            for i in range(s.x.shape[0]):
                for k in range(2):
                    spin_densities[j, i, k] = (
                        np.sum(abs(full[i, k, ...]) ** 2)
                        * s.dx ** (s.count - 1)
                        * s.count
                    )
        up_n = spin_densities[:, :, 0]
        down_n = spin_densities[:, :, 1]
        n = up_n + down_n
        if return_spins:
            return n, up_n, down_n
        else:
            return n

    if evolution is not None and type(evolution) == iDEA.state.SingleBodyEvolution:
        if time_indices is None:
            time_indices = np.array(range(evolution.t.shape[0]))
        up_n = np.zeros(shape=(time_indices.shape[0], s.x.shape[0]))
        for i, I in enumerate(evolution.up.occupied):
            for j, ti in enumerate(time_indices):
                up_n[j, :] += (
                    abs(evolution.up.td_orbitals[ti, :, i]) ** 2
                    * evolution.up.occupations[I]
                )
        down_n = np.zeros(shape=(time_indices.shape[0], s.x.shape[0]))
        for i, I in enumerate(evolution.down.occupied):
            for j, ti in enumerate(time_indices):
                down_n[j, :] += (
                    abs(evolution.down.td_orbitals[ti, :, i]) ** 2
                    * evolution.down.occupations[I]
                )
        n = up_n + down_n
        if return_spins:
            return n, up_n, down_n
        else:
            return n

    else:
        raise AttributeError(f"State or Evolution must be provided.")


def density_matrix(
    s: iDEA.system.System,
    state: Union[iDEA.state.SingleBodyState, iDEA.state.ManyBodyState] = None,
    evolution: Union[
        iDEA.state.SingleBodyEvolution, iDEA.state.ManyBodyEvolution
    ] = None,
    time_indices: np.ndarray = None,
    return_spins: bool = False,
) -> np.ndarray:
    r"""
    Compute the charge density matrix of a non_interacting state.

    | Args:
    |     s: iDEA.system.System, System object.
    |     state: iDEA.state.SingleBodyState or iDEA.state.ManyBodyState, State. (default = None)
    |     evolution: iDEA.state.SingleBodyEvolution or iDEA.state.ManyBodyEvolution, Evolution. (default = None)
    |     time_indices: np.ndarray, Time indices to compute observable if given evolution. If None will perform for all time indices. (default = None)
    |     return_spins: bool, True to also return the spin density matrices: total, up, down. (default = False)

    | Returns:
    |     p: np.ndarray, Charge density matrix, or evolution of charge density matrix.
    """
    if state is not None and type(state) == iDEA.state.ManyBodyState:
        tosum = list(range(2, s.count * 2))
        spin_p = (
            np.tensordot(state.full, state.full.conj(), axes=(tosum, tosum)).diagonal(
                axis1=1, axis2=3
            )
            * s.dx ** (s.count - 1)
            * s.count
        )
        up_p = spin_p[:, :, 0]
        down_p = spin_p[:, :, 1]
        p = up_p + down_p
        if return_spins:
            return p, up_p, down_p
        else:
            return p

    if state is not None and type(state) == iDEA.state.SingleBodyState:
        up_p = np.zeros(shape=s.x.shape * 2)
        down_p = np.zeros(shape=s.x.shape * 2)
        for i in range(state.up.orbitals.shape[1]):
            up_p += (
                np.tensordot(
                    state.up.orbitals[:, i], state.up.orbitals[:, i].conj(), axes=0
                )
                * state.up.occupations[i]
            )
        for i in range(state.down.orbitals.shape[1]):
            down_p += (
                np.tensordot(
                    state.down.orbitals[:, i], state.down.orbitals[:, i].conj(), axes=0
                )
                * state.down.occupations[i]
            )
        p = up_p + down_p
        if return_spins:
            return p, up_p, down_p
        else:
            return p

    if evolution is not None and type(evolution) == iDEA.state.ManyBodyEvolution:
        if time_indices is None:
            time_indices = np.array(range(evolution.t.shape[0]))
        tosum = list(range(2, s.count * 2))
        spin_density_matrices = np.zeros(
            shape=(time_indices.shape[0], s.x.shape[0], s.x.shape[0], 2), dtype=complex
        )
        for j, ti in enumerate(time_indices):
            l = string.ascii_lowercase[: s.count]
            L = string.ascii_uppercase[: s.count]
            st = (
                l
                + ","
                + L
                + "->"
                + "".join([i for sub in list(zip(l, L)) for i in sub])
            )
            full = np.einsum(st, evolution.td_space[ti, ...], evolution.spin)
            L = list(
                zip(list(range(0, s.count * 2, 2)), list(range(1, s.count * 2, 2)))
            )
            perms = itertools.permutations(list(range(s.count)))
            full_copy = copy.deepcopy(full)
            full = np.zeros_like(full_copy)
            for p in perms:
                indices = list(itertools.chain(*[L[e] for e in p]))
                full += iDEA.methods.interacting._permutation_parity(p) * np.moveaxis(
                    full_copy, list(range(s.count * 2)), indices
                )
            full = full / np.sqrt(np.sum(abs(full) ** 2) * s.dx**s.count)
            spin_density_matrices[j, :, :, :] = (
                np.tensordot(full, full.conj(), axes=(tosum, tosum)).diagonal(
                    axis1=1, axis2=3
                )
                * s.dx ** (s.count - 1)
                * s.count
            )
        up_p = spin_density_matrices[:, :, :, 0]
        down_p = spin_density_matrices[:, :, :, 1]
        p = up_p + down_p
        if return_spins:
            return p, up_p, down_p
        else:
            return p

    if evolution is not None and type(evolution) == iDEA.state.SingleBodyEvolution:
        if time_indices is None:
            time_indices = np.array(range(evolution.t.shape[0]))
        up_p = np.zeros(
            shape=(time_indices.shape[0], s.x.shape[0], s.x.shape[0]), dtype=complex
        )
        for i, I in enumerate(evolution.up.occupied):
            for j, ti in enumerate(time_indices):
                up_p[j, :] += (
                    np.tensordot(
                        evolution.up.td_orbitals[ti, :, i],
                        evolution.up.td_orbitals[ti, :, i].conj(),
                        axes=0,
                    )
                    * evolution.up.occupations[I]
                )
        down_p = np.zeros(
            shape=(time_indices.shape[0], s.x.shape[0], s.x.shape[0]), dtype=complex
        )
        for i, I in enumerate(evolution.down.occupied):
            for j, ti in enumerate(time_indices):
                down_p[j, :] += (
                    np.tensordot(
                        evolution.down.td_orbitals[ti, :, i],
                        evolution.down.td_orbitals[ti, :, i].conj(),
                        axes=0,
                    )
                    * evolution.down.occupations[I]
                )
        p = up_p + down_p
        if return_spins:
            return p, up_p, down_p
        else:
            return p

    else:
        raise AttributeError(f"State or Evolution must be provided.")


def kinetic_energy(
    s: iDEA.system.System,
    state: iDEA.state.SingleBodyState = None,
    evolution: iDEA.state.SingleBodyEvolution = None,
) -> Union[float, np.ndarray]:
    r"""
    Compute the kinetic energy of a non_interacting state.

    | Args:
    |     s: iDEA.system.System, System object.
    |     state: iDEA.state.SingleBodyState, State. (default = None)
    |     evolution: iDEA.state.SingleBodyEvolution, Evolution. (default = None)

    | Returns:
    |     E_k: float or np.ndarray, Kinetic energy, or evolution of kinetic energy.
    """
    if state is not None and type(state) == iDEA.state.ManyBodyState:
        K = iDEA.methods.interacting.kinetic_energy_operator(s)
        return observable(s, K, state=state)

    if state is not None and type(state) == iDEA.state.SingleBodyState:
        K = iDEA.methods.non_interacting.kinetic_energy_operator(s)
        return observable(s, K, state=state)

    if evolution is not None and type(evolution) == iDEA.state.ManyBodyEvolution:
        K = iDEA.methods.interacting.kinetic_energy_operator(s)
        return observable(s, K, evolution=evolution)

    if evolution is not None and type(evolution) == iDEA.state.SingleBodyEvolution:
        K = iDEA.methods.non_interacting.kinetic_energy_operator(s)
        return observable(s, K, evolution=evolution)

    else:
        raise AttributeError(f"State or Evolution must be provided.")


def external_potential(s: iDEA.system.System) -> np.ndarray:
    r"""
    Compute the external potential.

    | Args:
    |     s: iDEA.system.System, System object.

    | Returns:
    |     v_ext: np.ndarray, External potential of the system.
    """
    return s.v_ext


def external_energy(
    s: iDEA.system.System, n: np.ndarray, v_ext: np.ndarray
) -> Union[float, np.ndarray]:
    r"""
    Compute the external energy from a density and external potential.

    Args:
    |     s: iDEA.system.System, System object.
    |     n: np.ndarray, Charge density of the system.
    |     v_ext: np.ndarray, External potential of the system.

    Returns:
    |     E_ext: float or np.ndarray, External energy, or evolution of external energy.
    """
    if len(n.shape) == 1:
        E_ext = np.dot(n, v_ext) * s.dx
        return E_ext

    elif len(n.shape) == 2:
        E_ext = np.zeros(shape=n.shape[0])
        for j in range(E_ext.shape[0]):
            E_ext[j] = np.dot(n[j, :], v_ext[:]) * s.dx
        return E_ext

    else:
        raise AttributeError(f"Expected array of shape 1 or 2, got {n.shape} instead.")


def hartree_potential(s: iDEA.system.System, n: np.ndarray) -> np.ndarray:
    r"""
    Compute the Hartree potential from a density.

    | Args:
    |     s: iDEA.system.System, System object.
    |     n: np.ndarray, Charge density of the system.

    | Returns:
    |     v_h: np.ndarray, Hartree potential, or evolution of Hartree potential.
    """
    if len(n.shape) == 1:
        v_h = np.dot(n, s.v_int) * s.dx
        return v_h

    elif len(n.shape) == 2:
        v_h = np.zeros_like(n)
        for j in range(v_h.shape[0]):
            v_h[j, :] = np.dot(n[j, :], s.v_int[:, :]) * s.dx
        return v_h

    else:
        raise AttributeError(f"Expected array of shape 1 or 2, got {n.shape} instead.")


def hartree_energy(
    s: iDEA.system.System, n: np.ndarray, v_h: np.ndarray
) -> Union[float, np.ndarray]:
    r"""
    Compute the Hartree energy from a density and Hartree potential.

    | Args:
    |     s: iDEA.system.System, System object.
    |     n: np.ndarray, Charge density of the system.
    |     v_h: np.ndarray, Hartree potential of the system.


    | Returns:
    |     E_h: float or np.ndarray, Hartree energy, or evolution of Hartree energy.
    """
    if len(n.shape) == 1:
        E_h = 0.5 * np.dot(n, v_h) * s.dx
        return E_h

    elif len(n.shape) == 2:
        E_h = np.zeros(shape=n.shape[0])
        for j in range(E_h.shape[0]):
            E_h[j] = 0.5 * np.dot(n[j, :], v_h[j, :]) * s.dx
        return E_h

    else:
        raise AttributeError(f"Expected array of shape 1 or 2, got {n.shape} instead.")


def exchange_potential(s: iDEA.system.System, p: np.ndarray) -> np.ndarray:
    r"""
    Compute the exchange potential from a density matrix.

    | Args:
    |     s: iDEA.system.System, System object.
    |     p: np.ndarray, Density matrix of the system.

    | Returns:
    |     v_x: np.ndarray, Exchange potential, or evolution of exchange potential.
    """
    if len(p.shape) == 2:
        v_x = -p * s.v_int
        return v_x

    elif len(p.shape) == 3:
        v_x = np.zeros_like(p)
        for j in range(v_x.shape[0]):
            v_x[j, :, :] = -p[j, :, :] * s.v_int[:, :]
        return v_x

    else:
        raise AttributeError(f"Expected array of shape 1 or 2, got {n.shape} instead.")


def exchange_energy(
    s: iDEA.system.System, p: np.ndarray, v_x: np.ndarray
) -> Union[float, np.ndarray]:
    r"""
    Compute the exchange energy from a density matrix and exchange potential.

    | Args:
    |     s: iDEA.system.System, System object.
    |     p: np.ndarray, Density matrix of the system.
    |     v_x: np.ndarray, Exchange potential of the system.

    | Returns:
    |     E_x: float or np.ndarray, Exchange energy, or evolution of exchange energy.
    """
    if len(p.shape) == 2:
        E_x = 0.5 * np.tensordot(p, v_x, axes=2) * s.dx * s.dx
        return E_x

    elif len(p.shape) == 3:
        E_x = np.zeros(shape=p.shape[0], dtype=complex)
        for j in range(E_x.shape[0]):
            E_x[j] = 0.5 * np.tensordot(p[j, :, :], v_x[j, :, :], axes=2) * s.dx * s.dx
        return E_x.real

    else:
        raise AttributeError(f"Expected array of shape 1 or 2, got {n.shape} instead.")


def single_particle_energy(
    s: iDEA.system.System, state: iDEA.state.SingleBodyState
) -> float:
    r"""
    Compute the single particle energy of a single particle state.

    | Args:
    |     s: iDEA.system.System, System object.
    |     state: iDEA.state.SingleBodyState, State.

    | Returns:
    |     E: float, Single particle energy.
    """
    return np.sum(state.up.energies[:] * state.up.occupations[:]) + np.sum(
        state.down.energies[:] * state.down.occupations[:]
    )


def _placeholder(
    s: iDEA.system.System,
    state: Union[iDEA.state.SingleBodyState, iDEA.state.ManyBodyState] = None,
    evolution: Union[
        iDEA.state.SingleBodyEvolution, iDEA.state.ManyBodyEvolution
    ] = None,
    return_spins: bool = False,
) -> Union[float, np.ndarray]:

    r"""
    Placeholer function. Use this as a template when constructing observable methods.

    | Args:
    |     s: iDEA.system.System, System object.
    |     state: iDEA.state.SingleBodyState or iDEA.state.ManyBodyState, State. (default = None)
    |     evolution: iDEA.state.SingleBodyEvolution or iDEA.state.ManyBodyEvolution, Evolution. (default = None)
    |     return_spins: bool, True to also return the spin placeholer: total, up, down. (default = False)

    | Returns:
    |     observable: float or np.ndarray, Placeholer.
    """
    if state is not None and type(state) == iDEA.state.SingleBodyState:
        raise NotImplementedError()
    if state is not None and type(state) == iDEA.state.ManyBodyState:
        raise NotImplementedError()
    if evolution is not None and type(evolution) == iDEA.state.SingleBodyEvolution:
        raise NotImplementedError()
    if evolution is not None and type(evolution) == iDEA.state.ManyBodyEvolution:
        raise NotImplementedError()
    else:
        raise AttributeError(f"State or Evolution must be provided.")



import iDEA.utilities
import iDEA.system
import iDEA.interactions
import iDEA.state
import iDEA.observables
import iDEA.methods.interacting
import iDEA.methods.non_interacting
import iDEA.methods.hartree
import iDEA.methods.hartree_fock
import iDEA.methods.lda
import iDEA.methods.hybrid
import iDEA.reverse_engineering


__all__ = [
    "iDEA.utilities",
    "iDEA.system",
    "iDEA.interactions",
    "iDEA.state",
    "iDEA.observables",
    "iDEA.methods.interacting",
    "iDEA.methods.non_interacting",
    "iDEA.methods.hartree",
    "iDEA.methods.hartree_fock",
    "iDEA.methods.lda",
    "iDEA.methods.hybrid",
    "iDEA.reverse_engineering",
    "iterate_methods",
    "iterate_mb_methods",
    "iterate_sb_methods",
]


iterate_methods = [
    iDEA.methods.interacting,
    iDEA.methods.non_interacting,
    iDEA.methods.hartree,
    iDEA.methods.hartree_fock,
    iDEA.methods.lda,
    iDEA.methods.hybrid,
]
iterate_mb_methods = [iDEA.methods.interacting]
iterate_sb_methods = [
    iDEA.methods.non_interacting,
    iDEA.methods.hartree,
    iDEA.methods.hartree_fock,
    iDEA.methods.lda,
    iDEA.methods.hybrid,
]



"""Defines the structures to describe the system states"""


from abc import ABC as Interface
import copy
import numpy as np
import iDEA.utilities


__all__ = [
    "State",
    "ManyBodyState",
    "SingleBodyState",
    "Evolution",
    "ManyBodyEvolution",
    "SingleBodyEvolution",
]


class State(Interface):
    """Interface class representing a static state."""


class Evolution(Interface):
    """Interface class representing a time-dependent evolution of a state."""


class ManyBodyState(State):
    """State of interacting particles."""

    def __init__(
        self, space: np.ndarray = None, spin: np.ndarray = None, full=None, energy=None
    ):
        r"""
        State of particles in a many-body state.

        This is described by a spatial part 
        .. math:: \psi(x_1,x_2,\dots,x_N)
        on the spatial grid, and a spin
        part on the spin grid
        .. math:: \chi(\sigma_1,\sigma_2,\dots,\sigma_N).
        These are NOT necessarily antisymmetric states,
        they can be combined using the antisymmetrisation operaration to produce the full
        wavefunction
        .. math:: \Psi(x_1,\sigma_1,x_2,\sigma_2,\dots,x_N,\sigma_N).

        | Args:
        |     space: np.ndarray, Spatial part of the wavefunction on the spatial grid \psi(x_1,x_2,\dots,x_N). (default = None)
        |     spin: np.ndarray, Spin part of the wavefunction on the spin grid \chi(\sigma_1,\sigma_2,\dots,\sigma_N). (default = None)
        |     full: np.ndarray, Total antisymmetrised wavefunction \Psi(x_1,\sigma_1,x_2,\sigma_2,\dots,x_N,\sigma_N). (default = None)
        |     energy: float, Total energy of the state.
        """
        if space is None:
            self.space = iDEA.utilities.ArrayPlaceholder()
        else:
            self.space = space
        if spin is None:
            self.spin = iDEA.utilities.ArrayPlaceholder()
        else:
            self.spin = spin
        if full is None:
            self.full = iDEA.utilities.ArrayPlaceholder()
        else:
            self.full = full
        if energy is None:
            self.energy = float()
        else:
            self.energy = energy


class SingleBodyState(State):
    r"""
    State of particles in a single-body state.

    This is described by three arrays for each spin channel:

    | up.energies: np.ndarray, Array of single-body energies, indexed as energies[orbital_number].
    | up.orbitals: np.ndarray, Array of single-body orbitals, indexed as orbitals[space,orbital_number].
    | up.occupations: np.ndarray, Array of single-body occupations, indexed as occupations[orbital_number].
    | up.occupied: np.ndarray, Indices of up.occupations that are non-zero, to indicate occupied orbitals.

    | down.energies: np.ndarray, Array of single-body energies, indexed as energies[orbital_number].
    | down.orbitals: np.ndarray, Array of single-body orbitals, indexed as orbitals[space,orbital_number].
    | down.occupations: np.ndarray, Array of single-body occupations, indexed as occupations[orbital_number].
    | down.occupied: np.ndarray, Indices of down.occupations that are non-zero, to indicate occupied orbitals.
    """

    def __init__(self):
        self.up = iDEA.utilities.Container()
        self.down = iDEA.utilities.Container()

        self.up.energies = iDEA.utilities.ArrayPlaceholder()
        self.up.orbitals = iDEA.utilities.ArrayPlaceholder()
        self.up.occupations = iDEA.utilities.ArrayPlaceholder()
        self.up.occupied = iDEA.utilities.ArrayPlaceholder()

        self.down.energies = iDEA.utilities.ArrayPlaceholder()
        self.down.orbitals = iDEA.utilities.ArrayPlaceholder()
        self.down.occupations = iDEA.utilities.ArrayPlaceholder()
        self.down.occupied = iDEA.utilities.ArrayPlaceholder()


class ManyBodyEvolution(Evolution):
    r"""
    Time-dependent evolution of particles in a many-body state.

    In addition to the arrays defined within the initial ManyBodyState, this state is described by three additional arrays:

    | td_space: np.ndarray, Spatial part of the wavefunction on the spatial grid \psi(t,x_1,x_2,\dots,x_N).
    | v_ptrb: np.ndarray, Perturbation potential that this time-dependence was driven by. indexed as v_ptrb[space] if static, and v_ptrb[time,space] if dynamic.
    | t: np.ndarray, Time grid used during evolution.
    """

    def __init__(self, initial_state: ManyBodyState):
        self.space = copy.deepcopy(initial_state.space)
        self.spin = copy.deepcopy(initial_state.spin)
        self.full = copy.deepcopy(initial_state.full)
        self.td_space = iDEA.utilities.ArrayPlaceholder()
        self.v_ptrb = iDEA.utilities.ArrayPlaceholder()
        self.t = iDEA.utilities.ArrayPlaceholder()


class SingleBodyEvolution(Evolution):
    r"""
    Time-dependent evolution of particles in a single-body state.

    In addition to the arrays defined within the initial SingleBodyState, this state is described by four additional arrays:

    | up.td_orbitals: np.ndarray, Array of single-body time-dependend orbitals, indexed as orbitals[time,space,orbital_number].
    | down.td_orbital: np.ndarray, Array of single-body time-dependend orbitals, indexed as orbitals[time,space,orbital_number].
    | v_ptrb: np.ndarray, Perturbation potential that this time-dependence was driven by. indexed as v_ptrb[space] if static, and v_ptrb[time,space] if dynamic.
    | t: np.ndarray, Time grid used during evolution.

    In this case, only the occupied time-dependent orbitals are stored, as only these are propigated.
    """

    def __init__(self, initial_state: SingleBodyState):
        self.up = copy.deepcopy(initial_state.up)
        self.down = copy.deepcopy(initial_state.down)
        self.up.td_orbitals = iDEA.utilities.ArrayPlaceholder()
        self.down.td_orbitals = iDEA.utilities.ArrayPlaceholder()
        self.v_ptrb = iDEA.utilities.ArrayPlaceholder()
        self.t = iDEA.utilities.ArrayPlaceholder()
